#include <math.h>
#include <float.h>
#include <stdbool.h>
#include <stdio.h>
#include "gaussp.h"
#include "lustre.h"



long double RF(long double x, long double y, long double z)
{
 long double AmF = (x + y + z)/3, lambda, xm = x, ym = y, zm = z;
 long double X, Y, Z, E2, E3;
 long double AmxF = AmF - x, AmyF = AmF - y, AmzF = AmF - z, Q = ABS(AmxF) > ABS(AmyF) ? AmxF : AmyF;
 long double powinvfour = 1;
 Q = ABS(AmzF) > ABS(Q) ? ABS(AmzF) : ABS(Q);
 Q *= 0.8326831776556043 / RESID;
 do{
        lambda = sqrtl(xm * ym) + sqrtl(ym * zm) + sqrtl(xm * zm);
        AmF = (AmF + lambda) / 4;
        xm = (xm + lambda) / 4;
        ym = (ym + lambda) / 4;
        zm = (zm + lambda) / 4;
        powinvfour /= 4;
        Q /= 4;
 }while(Q > AmF);
 X = AmxF * powinvfour / AmF;
 Y = AmyF * powinvfour / AmF;
 Z = AmzF * powinvfour / AmF;
 E2 = X * Y - Z * Z;
 E3 = X * Y * Z;
 return 1 / sqrtl(AmF) * (1 - E2 / 10 + E3 / 14 + E2 * E2 / 24 - 3 * E2 * E3 / 44);
}

long double RJ(long double x, long double y, long double z, long double p)
{
 long double X, Y, Z, P, E2, E3, E4, E5, t;
 long double AmJ = (x + y + z + 2 * p)/5, xm = x, ym = y, zm = z, pm = p;
 long double AmxJ = AmJ - x, AmyJ = AmJ - y, AmzJ = AmJ - z, AmpJ = AmJ - p, Q = ABS(AmxJ) > ABS(AmyJ) ? AmxJ : AmyJ;
 long double sqx, sqy, sqz, sqp;
 long double sqdelta0 = (p - x) * (p - y) * (p - z);
 long double lambda, sqdelta8 = sqrtl(ABS(sqdelta0));
 int sgn = SIGN(sqdelta0);
 long double dm, powinvfour = 1, powertwo = (sgn < 0 ? 3 : 6) / (sgn == 0 ? 1 : sqdelta8), sumJ = 0;
 Q = ABS(AmzJ) > ABS(Q) ? ABS(AmzJ) : ABS(Q);
 Q = ABS(AmpJ) > Q ? ABS(AmpJ) : Q;
 Q *= 0.8908987181403393 / RESID;
 do{
        sqx = sqrtl(xm);  sqy = sqrtl(ym);  sqz = sqrtl(zm); sqp = sqrtl(pm);
        lambda = sqx * sqy + sqy * sqz + sqx * sqz;
        dm = (sqp + sqx) * (sqp + sqy) * (sqp + sqz);
        switch(sgn)
        {
         case 1:
           sumJ += powertwo * atan2l(sqdelta8, dm);
           break;
         case 0:
           sumJ += 6 * powinvfour / dm;
           break;
         default:
           t = 2 * sqdelta8 / (dm - sqdelta8);
           sumJ += (logl(1 + t)) * powertwo;           
        }
        AmJ = (AmJ + lambda) / 4;
        xm = (xm + lambda) / 4;
        ym = (ym + lambda) / 4;
        zm = (zm + lambda) / 4;
        pm = (pm + lambda) / 4;
        powinvfour /= 4;
        sqdelta8 /= 8;
        powertwo *= 2;
 }while(Q * powinvfour > AmJ);
 X = AmxJ * powinvfour / AmJ;
 Y = AmyJ * powinvfour / AmJ;
 Z = AmzJ * powinvfour / AmJ;
 P = (- X - Y - Z) / 2;
 E2 = X * Y + X * Z + Y * Z - 3 * P * P;
 E3 = X * Y * Z + 2 * E2 * P + 4 * P * P * P;
 E4 = (2 * X * Y * Z + E2 * P + 3 * P * P * P) * P;
 E5 = X * Y * Z * P * P;
 return sumJ + powinvfour / sqrtl(AmJ * AmJ * AmJ) * (1 - E2 * 3 / 14 + E3 / 6 + E2 * E2 * 9 / 88 - 3 * E4 / 22 -  9 * E2 * E3 / 52 + 3 * E5 / 26);
}

long double RD(long double x, long double y, long double z)
{
 long double AmD = (x + y + 3 * z)/5, lambda, xm = x, ym = y, zm = z;
 long double X, Y, Z, E2, E3, E4, E5;
 long double AmxD = AmD - x, AmyD = AmD - y, AmzD = AmD - z, Q = ABS(AmxD) > ABS(AmyD) ? AmxD : AmyD;
 long double powinvfour = 1, sumD = 0;
 Q = ABS(AmzD) > ABS(Q) ? ABS(AmzD) : ABS(Q);
 Q *= 0.8326831776556043l / RESID;
 do{
        lambda = sqrtl(xm * ym) + sqrtl(ym * zm) + sqrtl(xm * zm);
        sumD += powinvfour / (sqrtl(zm) * (zm + lambda));
        AmD = (AmD + lambda) / 4;
        xm = (xm + lambda) / 4;
        ym = (ym + lambda) / 4;
        zm = (zm + lambda) / 4;
        powinvfour /= 4;
        Q /= 4;
 }while(Q > AmD);
 X = AmxD * powinvfour / AmD;
 Y = AmyD * powinvfour / AmD;
 Z = -(X + Y) / 3;
 E2 = X * Y - 6 * Z * Z;
 E3 = (3 * X * Y - 8 * Z * Z) * Z;
 E4 = (X * Y - Z * Z) * Z * Z * 3;
 E5 = X * Y * Z * Z * Z;
 return 3 * sumD + powinvfour / sqrtl(AmD * AmD * AmD) * (1 - 3 * E2 / 14 + E3 / 6 + E2 * E2 * 9 / 88 - 3 * E4 / 22 - 9 / 52 * E2 * E3 + 3 / 26 * E5);
}

extern inline long double ReArcCos(long double x)
{
 if(x >= 1) return 0;
 if(x <= -1) return PI;
 return acosl(x);
}

long double Phi(long double delta, long double r)
{
 long double x;
 if(delta == 0 || r == 0) return 0;
 x = (delta * delta + r * r - 1) / delta / r / 2;
 if(x >= 1) return 1;
 if(x <= -1) return -1;
 return x;
}

long double L0(long double delta, long double r)
{
 if(r <= 0 || delta < 0)
  return 0;
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double psi1 = (delta * delta + 1 - r * r) / 2;
 long double psi2 = (delta * delta + r * r - 1) / 2 / r;
 if(delta == 0)
 {
  psi1 = (-SIGN(psi1) + 1) * PI_2;
  psi2 = (-SIGN(psi2) + 1) * PI_2;
 }
 else
 {
  psi1 = ReArcCos(psi1 / delta);
  psi2 = ReArcCos(psi2 / delta);
 }
 long double Q = (1 - rmindelta * rmindelta) * (rplusdelta * rplusdelta - 1);
 if(Q <= 0)
  Q = 0;
 else
  Q = sqrtl(Q);
 return psi1 + psi2 * r * r - Q / 2;
}

long double D1L0(long double delta, long double r)
{
 if(r <= 0 || delta <= 0)
  return 0;
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double Q = (1 - rmindelta * rmindelta) * (rplusdelta * rplusdelta - 1);
 if(Q <= 0)
  return 0;
 else
  Q = sqrtl(Q);
 return -Q / delta;
}

long double D2L0(long double delta, long double r)
{
 if(r <= 0 || delta < 0)
  return 0;
 long double psi2 = (delta * delta + r * r - 1) / 2 / r;
 if(delta == 0)
  psi2 = (-2 * SIGN(psi2) + 1) * PI_2;
 else
  psi2 = ReArcCos(psi2 / delta);
 return 2 * r * psi2;
}

long double D11L0(long double delta, long double r)
{
 long double q = (1 - (r - delta) * (r - delta)) * ((r + delta) * (r + delta) - 1);
 if(q <= 0)
  q = 0;
 else
  q = 1 / sqrtl(q);
 if(r <= 0 || delta <= 0)
  return 0;
 return -Phi(delta, r) * 2 * r / delta * (r * r - delta * delta - 1) * q;
}

long double D12L0(long double delta, long double r)
{
 long double q = (1 - (r - delta) * (r - delta)) * ((r + delta) * (r + delta) - 1);
 if(q <= 0)
  q = 0;
 else
  q = 1 / sqrtl(q);
 if(r <= 0 || delta <= 0)
  return 0;
 return -2 * r / delta * (1 - r * r + delta * delta) * q;
}

long double D22L0(long double delta, long double r)
{
 long double q = (1 - (r - delta) * (r - delta)) * ((r + delta) * (r + delta) - 1);
 if(q <= 0)
  q = 0;
 else
  q = 1 / sqrtl(q);
 if(r <= 0 || delta <= 0)
  return 0;
 long double psi2 = (delta * delta + r * r - 1) / 2 / r;
 if(delta == 0)
  psi2 = (-2 * SIGN(psi2) + 1) * PI_2;
 else
  psi2 = ReArcCos(psi2 / delta);
 return 2 * psi2 - 2 * (1 + r * r - delta * delta) * q;
}

long double L1(long double delta, long double r)
{
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double rmindeltaq = rmindelta * rmindelta, rdelta4 = 4 * delta * r, k, n, s, c, P, F, E, res = 0, sqq;
 if(delta == 0.5L && r == 0.5L)
     return 0.602753106752153308376436683315L;
 if (ABS(rmindelta) >= 1 || r <= 0 || delta < 0)
 {
  if(r > delta) res += 2.0943951023931954923L;
  if(r == delta) res += 1.0471975511965977462L;
  return res;
 }
 k = rdelta4 / (1 - rmindeltaq);
 s = (k < 1) ? 1 : 1 / sqrtl(k);
 c = (k < 1) ? 1 : 1 / k;
 if(r == delta)
  P = 0;
 else
  {
   n = -rdelta4 / rmindeltaq;
   P = s * RF(ABS(1 - c), ABS(1 - c * k), 1) + s * c * n * RJ(ABS(1 - c), ABS(1 - c * k), 1, 1 - c * n) / 3;
   P *= 2 * rplusdelta;
   P /= - 3 * rmindelta;
  }
 F = s * RF(ABS(1 - c), ABS(1 - c * k), 1);
 F *= (- 3 * rmindelta * rplusdelta + (1 - rmindeltaq) * (rplusdelta * rplusdelta - 1)) * 2;
 F /= 9;
 res = P - F;
 sqq = sqrtl(1 - rmindeltaq);
 res /= sqq;
 E = s * RF(ABS(1 - c), ABS(1 - c * k), 1) - s * c * k / 3 * RD(ABS(1 - c), ABS(1 - c * k), 1);
 E *= (7 * r * r + delta * delta - 4) * 2 * sqq;
 E /= 9;
 res += E;
 if(r > delta) res += 2.0943951023931954923L;
 if(r == delta) res += 1.0471975511965977462L;
 return res;
}


long double D1L1(long double delta, long double r)
{
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double rmindeltaq = rmindelta * rmindelta, rdelta4 = 4 * delta * r, k, s, c, F, E;
 if (ABS(rmindelta) >= 1 || r <= 0 || delta <= 0) return 0;
 k = rdelta4 / (1 - rmindeltaq);
 s = (k < 1) ? 1 : 1 / sqrtl(k);
 c = (k < 1) ? 1 : 1 / k;
 F = s * RF(ABS(1 - c), ABS(1 - c * k), 1);
 F *= 2 * (rplusdelta * rplusdelta - 1);
 E = s * RF(ABS(1 - c), ABS(1 - c * k), 1) - s * c * k / 3 * RD(ABS(1 - c), ABS(1 - c * k), 1);
 E *= 2 * (1- r * r - delta * delta);
 return -(E + F) * sqrtl(1 - rmindeltaq) / delta / 3;
}

long double D2L1(long double delta, long double r)
{
 long double rmindelta = r - delta;
 long double rmindeltaq = rmindelta * rmindelta, rdelta4 = 4 * delta * r, k, s, c, E;
 if (ABS(rmindelta) >= 1 || r <= 0 || delta < 0) return 0;
 k = rdelta4 / (1 - rmindeltaq);
 s = (k < 1) ? 1 : 1 / sqrtl(k);
 c = (k < 1) ? 1 : 1 / k;
 E = s * RF(ABS(1 - c), ABS(1 - c * k), 1) - s * c * k / 3 * RD(ABS(1 - c), ABS(1 - c * k), 1);
 return 4 * r * E * sqrtl(1 - rmindeltaq);
}

long double D11L1(long double delta, long double r)
{
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double rmindeltaq = rmindelta * rmindelta, rplusdeltaq = rplusdelta * rplusdelta, rdelta4 = 4 * delta * r, k, s, c, F, E, phi;
 long double qm = 1 - rmindeltaq, qp = rplusdeltaq - 1, q = qp * qm;
 if (ABS(rmindelta) >= 1 || r <= 0 || delta <= 0 || q == 0 || delta == 0) return 0;

 k = rdelta4 / (1 - rmindeltaq);
 s = (k < 1) ? 1 : 1 / sqrtl(k);
 c = (k < 1) ? 1 : 1 / k;
 if(qm < 0)
  E = 0;
 else
 {
  E = s * RF(ABS(1 - c), ABS(1 - c * k), 1) - s * c * k / 3 * RD(ABS(1 - c), ABS(1 - c * k), 1);
  E *= sqrtl(qm) * (2 * r * r - 4 * delta * delta - 2);
 }
 if(qm <= 0)
  F = 0;
 else
 {
  F = s * RF(ABS(1 - c), ABS(1 - c * k), 1);
  F *= (qm * qp + 3 * (r * r - delta * delta - 1) * (r * r + delta * delta - 1)) / sqrtl(qm);
 }
 if(q <= 0)
  phi = 0;
 else
 {
  phi = Phi(delta, r);
  phi *= 2 * r / delta * (r * r - delta * delta - 1) * sqrtl((1 - r * r - delta * delta + 2 * r * delta * phi) / q);
 }
  return (E + F) / delta / delta / 3 - phi;
}



long double L2(long double delta, long double r)
{
 if(r <= 0 || delta < 0)
  return 0;
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double psi1 = (delta * delta + 1 - r * r) / 2;
 long double psi2 = (delta * delta + r * r - 1) / 2 / r;
 if(delta == 0)
 {
  psi1 = (-2 * SIGN(psi1) + 1) * PI_2;
  psi2 = (-2 * SIGN(psi2) + 1) * PI_2;
 }
 else
 {
  psi1 = ReArcCos(psi1 / delta);
  psi2 = ReArcCos(psi2 / delta);
 }
 long double Q = (1 - rmindelta * rmindelta) * (rplusdelta * rplusdelta - 1);
 if(Q <= 0)
  Q = 0;
 else
  Q = sqrtl(Q);
 return  psi1 / 2 + psi2 * r * r * (2 * delta * delta + r * r) / 2 - Q * (delta * delta + 5 * r * r + 1) / 8;
}

long double D1L2(long double delta, long double r)
{
 if(r <= 0 || delta <= 0)
  return 0;
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double psi2 = (delta * delta + r * r - 1) / 2 / r;
 psi2 = ReArcCos(psi2 / delta);
 long double Q = (1 - rmindelta * rmindelta) * (rplusdelta * rplusdelta - 1);
 if(Q <= 0)
  Q = 0;
 else
  Q = sqrtl(Q);
 return 2 * psi2 * delta * r * r  - (delta * delta + r * r + 1) / 2 / delta * Q;
}

long double D2L2(long double delta, long double r)
{
 if(r <= 0 || delta < 0)
  return 0;
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double psi2 = (delta * delta + r * r - 1) / 2 / r;
 if(delta == 0)
  psi2 = (-2 * SIGN(psi2) + 1) * PI_2;
 else
  psi2 = ReArcCos(psi2 / delta);
 long double Q = (1 - rmindelta * rmindelta) * (rplusdelta * rplusdelta - 1);
 if(Q <= 0)
  Q = 0;
 else
  Q = sqrtl(Q);
 return 2 * r * ((delta * delta + r * r) * psi2 - Q) ;
}


long double D11L2(long double delta, long double r)
{
 if(r <= 0 || delta <= 0)
  return 0;
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double psi = ReArcCos((delta * delta + r * r - 1) / 2 / r / delta), phi = Phi(delta, r);
 long double Q = (1 - rmindelta * rmindelta) * (rplusdelta * rplusdelta - 1);
 if(Q <= 0)
  Q = 0;
 else
  Q = sqrtl(Q);
 psi *= 2 * r * r;
 psi += Q * (r * phi - 2 * delta) / delta;
 if(Q > 0)
  Q = 1 / Q;
 return psi - (r * r - delta * delta - 1) * Q * 2 * r * phi * (r * r + delta * delta - 2 * r * delta * phi) / delta;
}

long double D12L2(long double delta, long double r)
{
 if(r <= 0 || delta <= 0)
  return 0;
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double psi = ReArcCos((delta * delta + r * r - 1) / 2 / r / delta), phi = Phi(delta, r);
 long double Q = (1 - rmindelta * rmindelta) * (rplusdelta * rplusdelta - 1);
 if(Q <= 0)
  Q = 0;
 else
  Q = sqrtl(Q);
 psi *= 4 * delta * r;
 psi -= 2 * r * Q / delta;
 if(Q > 0)
  Q = 1 / Q;
 return psi + (r * r - delta * delta - 1) * Q * 2 * r * (r * r + delta * delta - 2 * r * delta * phi) / delta;
}

long double D22L2(long double delta, long double r)
{
 if(r <= 0 || delta <= 0)
  return 0;
 long double rplusdelta = r + delta, rmindelta = r - delta;
 long double psi = ReArcCos((delta * delta + r * r - 1) / 2 / r / delta), phi = Phi(delta, r);
 long double Q = (1 - rmindelta * rmindelta) * (rplusdelta * rplusdelta - 1);
 if(Q <= 0)
  Q = 0;
 else
  Q = sqrtl(Q);
 psi *= 2 * (delta * delta + 3 * r * r);
 psi -= 4 * Q;
 if(Q > 0)
  Q = 1 / Q;
 return psi + 2 * (delta * delta - r * r - 1) * Q * (r * r + delta * delta - 2 * r * delta * phi);
}